// BUSINESS LOGIC

// File contains code from Jumpcore; notice applies to that code only:
/* Copyright (C) 2008-2009 Andrew McClure
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// File contains code from Chipmunk "MoonBuggy" demo; notice applies to that code only:
/* Copyright (c) 2007 Scott Lembcke
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "chipmunk.h"
#include "color.h"
#include "program.h"

#define GRAVITY cpv(0.0f, -9.8/2)

vector<spaceinfo> spaces;

bool wantClearUi = false;
WantNewScreen want = WNothing;

// /------ Used by the typewriter demo ------\

bool optDebug = false, optCursor = true;
cpBody *nextWord = NULL;
cpVect wordOrigin = cpv(-0.8,0.8);
cpBody *mouseBody = NULL;
cpJoint *mouseJoint = NULL;
cpBody *allocWord() {
	cpBody *b = cpBodyNew(INFINITY, INFINITY);
	b->data = new wordinfo(false);
	return b;
}

CheckControl *debugBox, *screenBox, *cursorBox;
class OptionsControl : public ControlBase { public:
	OptionsControl(string _text = string(), bool _bg = false) : ControlBase(_text,_bg) {}
	virtual void click() {
		optDebug = debugBox->checked;
		optCursor = cursorBox->checked;
		
		if (optWindow != (!screenBox->checked)) { // Only call recreate_surface if screenBox value has changed
			optWindow = !optWindow;
			recreate_surface(optWindow);
		}
		
		wantClearUi = true;
	}
};

cpVect lastClick = cpvzero;
cpShape *draggingMe = NULL;
cpVect draggingMeInto = cpvzero;
unsigned int lastLetterAt = 0;

void mouseDownOnShape(cpShape *shape, void *data) {
	draggingMe = shape;
	draggingMeInto = cpvunrotate( cpvsub(lastClick, draggingMe->body->p), draggingMe->body->rot );
}

// \------ begin display state ------/

void appendLevel(double depth) {
	int s = spaces.size();
  /* We first create a new space */
  spaces.push_back(spaceinfo());
  spaces[s].depth = depth;  
  
	spaces[s].space = cpSpaceNew();
	
	/* Next, you'll want to set the properties of the space such as the
     number of iterations to use in the constraint solver, the amount
     of gravity, or the amount of damping. In this case, we'll just
     set the gravity. */
	spaces[s].space->gravity = GRAVITY;
	
	/* This step is optional. While you don't have to resize the spatial
	 hashes, doing so can greatly increase the speed of the collision
	 detection. The first number should be the expected average size of
	 the objects you are going to have, the second number is related to
	 the number of objects you are putting. In general, if you have more
	 objects, you want the number to be bigger, but only to a
	 point. Finding good numbers to use here is largely going to be guess
	 and check. */
	cpSpaceResizeStaticHash(spaces[s].space, 0.1, 2000);
	cpSpaceResizeActiveHash(spaces[s].space, 0.1, 200);
	
  /* This is the rigid body that we will be attaching our ground line
     segments to. We don't want it to move, so we give it an infinite
     mass and moment of inertia. We also aren't going to add it to our
     space. If we did, it would fall under the influence of gravity,
     and the ground would fall with it. */
  spaces[s].staticBody = cpBodyNew(INFINITY, INFINITY);
}

// This is called once when the program begins. Clear out and replace with your own code
void
program_init(void)
{
	nextWord = allocWord();
	nextWord->p = wordOrigin;

	appendLevel(0);
	
	cpVect screen_verts[] = {
		cpv(-1/aspect,-1),
		cpv(-1/aspect, 1),
		cpv( 1/aspect, 1),
		cpv( 1/aspect,-1),
	};
	for(int c = 0; c < 4; c++) {		
		cpShape *newShape = cpSegmentShapeNew(spaces[0].staticBody, screen_verts[c], screen_verts[(c+1)%4], 0.1);
		cpSpaceAddStaticShape(spaces[0].space, newShape);
		newShape->e = 0.8; newShape->u = 0.8;
	}

	mouseBody = cpBodyNew(INFINITY, INFINITY);
	//cpShape *addme = cpCircleShapeNew(mouseBody,  0.1, cpvzero); // ?
	cpSpaceAddBody(spaces[0].space, mouseBody);
	//cpSpaceAddShape(spaces[0].space, addme); // ?
}

// The default display() calls this once per framedraw.
void
program_update(void)
{
	if (paused) return; // No motion when paused
	
  /* Collision detection isn't amazing in Chipmunk yet. Fast moving
     objects can pass right though each other if they move to much in a
     single step. To deal with this, you can just make your steps
     smaller and cpSpaceStep() sevral times. */
  int substeps = 3;

  /* This is the actual time step that we will use. */
  // Note: Chipmunk will not perform as well unless the step stays fixed over time
#if !DYNAMIC_FRAMERATE
  cpFloat dt = (1.0f/FPS) / (cpFloat)substeps;
#else
  cpFloat dt = sinceLastFrame / 1000.0 / (cpFloat)substeps;
#endif
	
  int i;
  for(i=0; i<substeps; i++){
	  
	  if (draggingMe) { // Typewriter demo
		  cpVect force = cpvmult(cpvsub(lastClick, mouseBody->p), FPS);
//		  ERR("V %f, %f\n", mouseBody->p.x, mouseBody->p.y, force.x, force.y);
		  mouseBody->v = force;		  
	  }
	  
	  for(vector<spaceinfo>::iterator _s = spaces.begin(); _s != spaces.end(); _s++) {
		spaceinfo *s = &(*_s);
		if (okayToDraw(s)) {		
			// Iterate space normally
			cpSpaceStep(s->space, dt);
		}
	  }
	}
}

// This is called when SDL gets a key event. Clear out and replace with your own code
void program_eventkey(SDL_Event &event) {
	if (event.type == SDL_KEYDOWN) {
		tapping = TAPFULL;
		lastLetterAt = ticks;
		
		wordinfo *current = (wordinfo *)nextWord->data;
		
		SDLKey &key = event.key.keysym.sym;

		if (key == SDLK_F1) {
			if (0 == columns.size()) {
				wantClearUi = true;
				want = WOptions;
				program_interface();
			}
		} else if (key == SDLK_BACKSPACE) {
			if (current->text.size() > 0)
				current->text.erase(current->text.size()-1);
		} else if (key == SDLK_RETURN || key == SDLK_TAB || key == SDLK_SPACE) {
			if (current->text.length() > 0) {
				float width = textWidth(current->text)/2;
				float height = textHeight(current->text)/2;
				cpVect slice_verts[] = {
					cpv(-width,-height),
					cpv(-width, height),
					cpv( width, height),
					cpv( width,-height),
				};
				
				cpBody *newWord = allocWord();
				cpShape *newShape = cpPolyShapeNew(newWord, 4, slice_verts, cpvzero);
				wordinfo *word = (wordinfo *)newWord->data;
				
				newWord->p = cpv(nextWord->p.x+width,nextWord->p.y+height);
				newShape->e = 0.80;
				newShape->u = 0.80;
				newShape->collision_type = C_WORD;
				cpBodySetMass(newWord, 5.0*current->text.length());
				cpBodySetMoment(newWord, cpMomentForPoly(newWord->m, 4, slice_verts, cpvzero));
				word->text = current->text;
				word->center = true;
				
				cpSpaceAddBody(spaces[0].space, newWord);
				cpSpaceAddShape(spaces[0].space, newShape);
			}
				
			if (key != SDLK_RETURN) {
				if (key == SDLK_SPACE)
					current->text += " ";
				else
					current->text += "        ";
				
				if (current->text.length() > 1)
					nextWord->p.x += textWidth(current->text);
				else
					nextWord->p.x += textWidth("  ")/2; // There's some kinda weird bug where textWidth(" ") comes back 0
				current->text = "";
			}
			
			if (key == SDLK_RETURN || nextWord->p.x > -wordOrigin.x) {
				nextWord->p.y -= textHeight(current->text);
				nextWord->p.x = wordOrigin.x;
				current->text = "";
				
				if (nextWord->p.y < -wordOrigin.y) {
					nextWord->p.y = wordOrigin.y;
				}
			}
		} else if (event.key.keysym.unicode >= 32 && event.key.keysym.unicode <= 126)
			current->text += (char)event.key.keysym.unicode;
	}
}

// This is called when SDL gets a joystick event.
void program_eventjoy(SDL_Event &event) {
}

// This is called when SDL gets a mouse event.
void program_eventmouse(SDL_Event &event) {
	if (event.type == SDL_MOUSEBUTTONUP && draggingMe) {
		cpBodyResetForces(draggingMe->body);
		draggingMe = NULL;
		cpSpaceRemoveJoint(spaces[0].space, mouseJoint);
		cpJointDestroy(mouseJoint);
		return;
	}
	goOrtho();
	glLoadIdentity();
	lastClick = screenToGL(event.button.x, event.button.y, spaces[0].depth);
	if (event.type == SDL_MOUSEBUTTONDOWN) {
		draggingMe = NULL;
		cpSpaceShapePointQuery(spaces[0].space, lastClick, mouseDownOnShape, NULL);
		
		if (draggingMe) {
			mouseBody->p = lastClick;
			mouseJoint = cpPivotJointNew(draggingMe->body, mouseBody, lastClick);
			cpSpaceAddJoint(spaces[0].space, mouseJoint);
		}
	}
}

// The "this is a good time to modify your uiSpace" function. Main.cpp calls this once at the beginning of the main
// loop and once every time a button is successfully clicked. This exists because it's unsafe to modify the interface
// from inside of a ControlBase click() handler, so instead click() handlers can stash some state that this function
// then picks up when main.cpp automatically calls it after the click is completed...
void program_interface() {
	WantNewScreen wantAfterwards = WNothing; // "want" will be cleared with this at the end of the function
	
	if (wantClearUi) {
		for(int c = 0; c < columns.size(); c++)
			if (columns[c])
				delete columns[c];
		
		columns.clear();
		
		resetScrollmax();
		wantClearUi = false;
	}
	
	switch (want) {	
		case WOptions: {
			ContainerBase *checkboxes = new ContainerBase(uiSpace, CMID);
			columns.push_back(checkboxes);

			checkboxes->add( new ControlBase("Options") );
			checkboxes->add( new ControlBase("") );
			
			screenBox = new CheckControl("Full\nScreen", true);
			if (!optWindow)
				screenBox->click();
			checkboxes->add( screenBox );
			cursorBox = new CheckControl("Cursor", true);
			if (optCursor)
				cursorBox->click();
			checkboxes->add( cursorBox );
			debugBox = new CheckControl("Draw\nDebug", true);
			if (optDebug)
				debugBox->click();
			checkboxes->add( debugBox );
			
			checkboxes->add( new ControlBase("") );
			checkboxes->add( new OptionsControl("Save", true) );
			checkboxes->commit();
		} break;
	}
	
	want = wantAfterwards;
}

// Called when "ESC" is pressed-- gives you a last chance to do something like throw up an "are you sure?" screen
void BackOut() {
	Quit(); // Default implementation just quits.
}

// Called after SDL quits, right before program terminates. Gives you a chance to save to disk anything you want to save.
void AboutToQuit() {
	// Default implementation does nothing.
}